  /* $Workfile: T2p5x4Recorder.c $       $Revision: 32 $      $Modtime: 1/02/07 6:31p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T2P5X4RECORDER.C

    A simple example console application demonstrating the use of Raw Data Sender
    and the Raw Data Receiver OTX Driver.
    
   The applications works with the following board combinations:

      Thor-2-PCI / Thor-8-PCI + Vidar-5x4-ASM,

    or OTX PCI Plus series boards,

      Thor-2-PCI-Plus / Thor-4-PCI-Plus / Thor-8-PCI-Plus + Vidar-5x4-ASM-CST,
      Thor-2-PCI-Plus / Thor-4-PCI-Plus / Thor-8-PCI-Plus + Vidar-5x4-ASM-PRO,
      Thor-2-PCI-Plus / Thor-4-PCI-Plus / Thor-8-PCI-Plus + Vidar-5x4-ASM-EX,
      Thor-2-PCI-Plus / Thor-4-PCI-Plus / Thor-8-PCI-Plus + Vidar-55x4-ASM,

    or OTX PCI Express series boards,

      Thor-2-PCI-Express

  The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    The Codec (HS0) is connected with two pipes to DSP 0.

    With keystroke command you can record voice from the HS microphone to the disk.
    You can also playback an already recorded voice file.

    A command line switch specifies whether U-law or A-law should be used for the
    recording/playback

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>


#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxDsp.h"
#include "OtxRaw.h"
#include "OtxT1E1.h"
#include "OtxData.h"

#define DEMO_TASK_REF 1234               // Task Reference for a non-blocking call


#define DEMO_RAW_TX_BUF_SIZE 1024


#define DEMO_DEFAULT_LOG_FILE "T2p5x4Recorder.log"
#define DEMO_DEFAULT_BOARD_NO 0


typedef struct DemoEnvT {
  char            *pszInputStr;
  OTX_BOOL        bExitOnError;
  char            *logFileName;
  OTX_INT32       nBoardType;
  OtxDataFormatE  eLaw;
  int             boardNo;
  int             coDsp;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;



struct {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[2]; // Only used for Thor-4-PCI-Plus, Thor-8-PCI(-Plus); not for Thor-2-PCI
  OTX_HANDLE hVidarAsmController0;
  OTX_HANDLE hDsp;
  OTX_HANDLE hCodec;
} g_hPhysDevs;


struct {
  OTX_HANDLE hDspCodec;
  OTX_HANDLE hCodecDsp;
} g_hPipes;

struct sRawDataSender {
  OTX_HANDLE hHandle;
  OTX_HANDLE hEventQueue;
  OTX_EVENT  hNotificationEvent;

  OTX_UINT32 nThreadId;
  OTX_HANDLE hThread;

  OTX_UINT8   sqTxBuf[DEMO_RAW_TX_BUF_SIZE];
  FILE       *pTxFile;
  OTX_INT32   cbTxBuf;
  OTX_INT32   nTaskRef;

};

struct sRawDataReceiver {
  OTX_HANDLE hHandle;
  OTX_HANDLE hEventQueue;
  OTX_EVENT  hNotificationEvent;

  OTX_UINT32 nThreadId;
  OTX_HANDLE hThread;

  FILE       *pRxFile;
  clock_t     nMsDuration;
};


struct {
  struct sRawDataSender mRawDataSender;
  struct sRawDataReceiver mRawDataReceiver;
} g_hLogicDevs;


/*******************************************************************************
*******************************************************************************/


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = NULL;    // D16V86 apps cannot use event objects
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the Thor board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;

  // Connect to the Physical Thor Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCI:
    DemoLogf("Opening OTX_DEVICE_THOR_PCI board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI:
    DemoLogf("Opening Thor-8-PCI board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_PLUS:
    DemoLogf("Opening Thor-2-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    DemoLogf("Opening Thor-2-PCI-Express board\n");
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    DemoLogf("Opening Thor-4-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf("Opening Thor-8-PCI-Plus board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // Get a handle to the Codec on the Thor-2-PCI board
  DemoLogf("Open OTX_DEVICE_CODEC\n");
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_CODEC, 0, 0, g_hEventQueue, &g_hPhysDevs.hCodec); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // Set PCM compression law for the Codec
  DemoLogf("Set Law to %s\n", g_demoEnv.eLaw == OTX_DATA_FORMAT_COMPAND8_A_LAW ? "A-law" : "U-law");
  result = OtxBrdCodecSetLaw(g_hPhysDevs.hCodec, g_demoEnv.eLaw); DemoCheckRc(g_hPhysDevs.hCodec, result);

  // Get a handle to to the ASM Daughter Board
  {
    OTX_UINT32   eDevType;
    OTX_ITERATOR itrChildTypes;
    OtxDrvIteratorBegin(&itrChildTypes);
    while (OTX_SUCCESS(OtxDrvEnumDeviceType(g_hPhysDevs.hThorController0, &itrChildTypes, (OTX_INT32 *)&eDevType))) {  
      OTX_BOOL bOpenAsm = OTX_FALSE;
      switch (eDevType) {
      case OTX_DEVICE_VIDAR_5x_ASM:
        DemoLogf("Opening Vidar-5x4-ASM board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_5X4_ASM_CST:
        DemoLogf("Opening Vidar-5x4-ASM-CST board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_5X4_ASM_V90:
        DemoLogf("Opening Vidar-5x4-ASM-V90 board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_5X4_ASM_EX:
        DemoLogf("Opening Vidar-5x4-ASM-EX board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_5X4_ASM_PRO:
        DemoLogf("Opening Vidar-5x4-ASM-PRO board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_55x4_ASM:
        DemoLogf("Opening Vidar-55x4-ASM board\n");
        bOpenAsm = OTX_TRUE;

        // Verify that the system supports the Vidar-55x4-ASM board
        {
          char szCpldBuildNumbers[256];
          char szBrdBuildNumber[16];
          OTX_INT32 nBoardBuildNo;
          
          if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCI) {
            
            //  Verify firmware revision
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, szCpldBuildNumbers, 256); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            if (strstr(szCpldBuildNumbers, "OTXCDC: 0761") || strstr(szCpldBuildNumbers, "OTXCDC: 0762")) {
              DemoLogf("Error: The firmware on the Thor-2-PCI must be upgraded to firmware build 5\n");
              DemoLogf("       (OTXCDC 0764) or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please download firmware update program from the Odin TeleSystems\n");
              DemoLogf("       web site (www.odints.com)\n");
              DemoLogf("       Also please verify that the HMA-1098-1-2.0 module is plugged into connector BJ3.\n");
              DemoExitQuestion();
            }
            
            // Verify board build number
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, szBrdBuildNumber, 16); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            nBoardBuildNo = strtol(strtok(szBrdBuildNumber, "."), NULL, 0);
            DemoLogf("Board Build Number: %d\n", nBoardBuildNo);
            if (nBoardBuildNo < 11) {
              DemoLogf("Error: The Thor-2-PCI board must be upgraded to build 11\n");
              DemoLogf("       or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please contact Odin TeleSystems Support Team for upgrade details.\n");
              DemoExitQuestion();
            }
            
          } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) {
            //  Verify firmware revision
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, szCpldBuildNumbers, 256); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            if (strstr(szCpldBuildNumbers, "OTXCDC: 0771")) {
              DemoLogf("Error: The firmware on the Thor-8-PCI must be upgraded to firmware build 3\n");
              DemoLogf("       (OTXCDC 0772) or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please download firmware update program from the Odin TeleSystems\n");
              DemoLogf("       web site (www.odints.com)\n");
              DemoLogf("       Also please verify that the HMA-1098-1-2.0 module is plugged into connector BJ3.\n");
              DemoExitQuestion();
            }
            
            // Verify board build number
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, szBrdBuildNumber, 16); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            nBoardBuildNo = strtol(szBrdBuildNumber, NULL, 0);
            DemoLogf("Board Build Number: %d\n", nBoardBuildNo);
            if (nBoardBuildNo < 5) {
              DemoLogf("Error: The Thor-8-PCI board must be upgraded to build 5\n");
              DemoLogf("       or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please contact Odin TeleSystems Support Team for upgrade details.\n");
              DemoExitQuestion();
            }
            
          }
          
        }
        break;
      }
      if (bOpenAsm) {
        // Get a handle to to the ASM Daughter Board
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, (OtxPhysicalDeviceTypeE) eDevType, 0, 0, g_hEventQueue, &g_hPhysDevs.hVidarAsmController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }      
    }
  }

  // Get a handle to the first DSP on the ASM Daugther board
  DemoLogf("Open OTX_DEVICE_DSP #0\n");
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hVidarAsmController0, OTX_DEVICE_DSP, 0, 0, g_hEventQueue, &g_hPhysDevs.hDsp); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);

	// Load DSP Program
  DemoLogf("Loading OTX_DSP_SPM_OTXHDLC to DSP#0\n");
  result = OtxDspRunSpm(g_hPhysDevs.hDsp, OTX_DSP_SPM_OTXHDLC); DemoCheckRc(g_hPhysDevs.hDsp, result);


}



/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;

  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  result = OtxDrvEnable(g_hPhysDevs.hVidarAsmController0, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);

  result = OtxDrvEnable(g_hPhysDevs.hDsp,            OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp, result);
  result = OtxDrvEnable(g_hPhysDevs.hCodec,          OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hCodec, result);
}



/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;

  // Create a notification Event and Event Queue for the Raw Data Sender
  result = OtxDrvCreateEvent(&g_hLogicDevs.mRawDataSender.hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  result = OtxDrvCreateEventQueue(128, g_hLogicDevs.mRawDataSender.hNotificationEvent, &g_hLogicDevs.mRawDataSender.hEventQueue); DemoCheckRc(g_hLogicDevs.mRawDataSender.hEventQueue, result);
  result = OtxDrvEnable(g_hLogicDevs.mRawDataSender.hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataSender.hEventQueue, result);
  
  // Create Raw Data Sender
  result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp, OTX_LDEVICE_DATA_RAW_SENDER, g_hLogicDevs.mRawDataSender.hEventQueue, 0, &g_hLogicDevs.mRawDataSender.hHandle); DemoCheckRc(g_hPhysDevs.hDsp, result);
  
  // Create a notification Event and Event Queue for the Raw Data Receiver
  result = OtxDrvCreateEvent(&g_hLogicDevs.mRawDataReceiver.hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  result = OtxDrvCreateEventQueue(128, g_hLogicDevs.mRawDataReceiver.hNotificationEvent, &g_hLogicDevs.mRawDataReceiver.hEventQueue); DemoCheckRc(g_hLogicDevs.mRawDataReceiver.hEventQueue, result);
  result = OtxDrvEnable(g_hLogicDevs.mRawDataReceiver.hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver.hEventQueue, result);
  
  // Create Raw Data Receiver
  result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp, OTX_LDEVICE_DATA_RAW_RECEIVER, g_hLogicDevs.mRawDataReceiver.hEventQueue, 0, &g_hLogicDevs.mRawDataReceiver.hHandle); DemoCheckRc(g_hPhysDevs.hDsp, result);

}



/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_RESULT result;

  result = OtxDrvCreatePipe(64, &g_hPipes.hCodecDsp); DemoCheckRc(g_hPipes.hCodecDsp, result);
  result = OtxDrvCreatePipe(64, &g_hPipes.hDspCodec); DemoCheckRc(g_hPipes.hDspCodec, result);
  
  result = OtxDrvConnectPipe(g_hPipes.hCodecDsp, OTX_PIPE_INPUT,  g_hPhysDevs.hCodec, -1, NULL, 0); DemoCheckRc(g_hPipes.hCodecDsp, result);
  result = OtxDrvConnectPipe(g_hPipes.hCodecDsp, OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp,   -1, NULL, 0); DemoCheckRc(g_hPipes.hCodecDsp, result);
  
  result = OtxDrvConnectPipe(g_hPipes.hDspCodec, OTX_PIPE_INPUT,  g_hPhysDevs.hDsp,   -1, NULL, 0); DemoCheckRc(g_hPipes.hDspCodec, result);
  result = OtxDrvConnectPipe(g_hPipes.hDspCodec, OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec, -1, NULL, 0); DemoCheckRc(g_hPipes.hDspCodec, result);
  
  // Create the following Connections in receive direction:
  //  Codec  ------> Raw Data Receiver
  result = OtxDrvConnectLogicalDevice(g_hPipes.hCodecDsp, -1, g_hLogicDevs.mRawDataReceiver.hHandle,       -1); DemoCheckRc(g_hPipes.hCodecDsp, result);
  
  // Make the following connections in transmit direction:
  // Raw Data Sender   -----> Codec
  result = OtxDrvConnectLogicalDevice(g_hLogicDevs.mRawDataSender.hHandle, -1, g_hPipes.hDspCodec, -1); DemoCheckRc(g_hLogicDevs.mRawDataSender.hHandle, result);

}


/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  OTX_RESULT result;

  // Initialize the Raw Data Receivers (hosted in the DSP)
  result = OtxRawDataReceiverInit(
    g_hLogicDevs.mRawDataReceiver.hHandle,
    OTX_DATA_FORMAT_UINT8,  // No data conversion. OtxHdlc SPM only supports this format
    0xFF,        // Flag to be filtered
    OTX_FALSE,   // Disable Flag filtering
    OTX_TRUE     // Least Significant part is received first. Only applicable for a subchannel pipe (nCapacity < 64).
    ); DemoCheckRc(g_hLogicDevs.mRawDataReceiver.hHandle, result);
  // Initialize the Raw Data Senders (hosted in the DSP)
  result = OtxRawDataSenderInit(
    g_hLogicDevs.mRawDataSender.hHandle, 
    OTX_DATA_FORMAT_UINT8,  // No data conversion. OtxHdlc SPM only supports this format
    0xFF,        // Flag to be sent when the sender is out of data
    OTX_TRUE     // Least Significant part is transmitted first. Only applicable for a subchannel pipe (nCapacity < 64).
    ); DemoCheckRc(g_hLogicDevs.mRawDataSender.hHandle, result);
  
}


/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
void DemoEnableLogicalDevices(void)
{
  OTX_RESULT result;

  // Enable Raw Data Sender (always on)
  result = OtxDrvEnable(g_hLogicDevs.mRawDataSender.hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataSender.hHandle, result);

  // We enable the Raw Data Receiver only when it are in use (to avoid interrupting the host 

}





/*******************************************************************************
  @func
    Raw Data Receiver Thread Function 
*******************************************************************************/
void *DemoRawDataReceiver(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;
  OTX_RESULT otxRc;
  OTX_UINT8  buf[2048];
  OTX_UINT32 nFrameLen;
  OTX_UINT32 nMaxFrameLen = 0;
  OTX_UINT32 nEventCount = 0;
  OTX_UINT32 nReadDataCount = 0;
  char msgStr[256];
  clock_t startTime, stopTime;

  startTime = clock();

  result = OtxDrvEnable(g_hLogicDevs.mRawDataReceiver.hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver.hHandle, result);


  while (clock() < (startTime + g_hLogicDevs.mRawDataReceiver.nMsDuration)) {
    OtxDrvWaitForSingleEvent(g_hLogicDevs.mRawDataReceiver.hNotificationEvent, OTX_TIME_INFINITY);
    do {
      result = OtxDrvGetEventData(g_hLogicDevs.mRawDataReceiver.hEventQueue, &eventData); DemoCheckRc(g_hLogicDevs.mRawDataReceiver.hEventQueue, result);

      nEventCount++;

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        switch (eventData.m_eDeviceType) {
        case OTX_LDEVICE_DATA_RAW_RECEIVER:
          // Event from Raw Data Receiver 

          switch (eventData.m_nCode) {
          case OTX_RAWDATA_RECEIVER_EC_DATA_AVAILABLE:
            {
              OTX_BOOL bFrameEnd;
              otxRc = OtxRawDataReadData(eventData.m_hDevice, buf, 2048, &nFrameLen, &bFrameEnd);
              while (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {

                nReadDataCount++;
                
                if (g_hLogicDevs.mRawDataReceiver.pRxFile) {
                  int cbWritten;

                  if (nFrameLen > nMaxFrameLen) {
                    nMaxFrameLen = nFrameLen;
                  }

                  cbWritten = fwrite(buf, sizeof(OTX_UINT8), nFrameLen, g_hLogicDevs.mRawDataReceiver.pRxFile);

                  if (cbWritten != (int)nFrameLen) {
                    DemoLogf("RawDataReceiver[%d] Error: Could not write data to file.  (nFrameLen=%d, cbWritten=%d) \n", eventData.m_nSourceId, nFrameLen, cbWritten);
                  }
                  
                  if (bFrameEnd) {
                    DemoLogf("RawDataSender[%d] Frame End occurred (unexpectedly). Data reception complete. Closing data file. \n", eventData.m_nSourceId);
                    fclose(g_hLogicDevs.mRawDataReceiver.pRxFile);
                    g_hLogicDevs.mRawDataReceiver.pRxFile = NULL;
                    goto END;
                  }
                }
                
                otxRc = OtxRawDataReadData(eventData.m_hDevice, buf, 2048, &nFrameLen, &bFrameEnd);
              }
              if (OTX_RESULT_CODE(otxRc) != OTX_S_FALSE) {
                DemoLogf("RawDataReceiver[%d] OtxRawDataReadData() was unsuccessful.\n", eventData.m_nSourceId);
                DemoCheckRc(eventData.m_hDevice, otxRc);
              }

            }
            break;
          default:
            otxRc = OtxDrvEventCode2String(eventData.m_hDevice, eventData.m_nCode, 256, &msgStr[0]);
            // Unexpected event from this device
            if (OTX_SUCCESS(otxRc)) {
              DemoLogf("RawDataReceiver[%d] Event: %s \n", eventData.m_nSourceId, msgStr);
            } else {
              DemoLogf("RawDataReceiver[%d] Invalid Event: %ld \n", eventData.m_nSourceId, eventData.m_nCode);
            }
            break;
          }
          break;
          default:
            // Unhandled event
            DemoLogf("DemoRawDataSender: Event from an unknown object 0x%X\n", eventData.m_hDevice);
            break;
        }
        
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

    OsaThreadSleep(256); // Give the driver some time to fill up more data for us to fetch (256ms fills 2048 bytes in a 64kbit/s pipe)

  }

END:
  
  stopTime = clock();

  result = OtxDrvDisable(g_hLogicDevs.mRawDataReceiver.hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver.hHandle, result);

  if (g_hLogicDevs.mRawDataReceiver.pRxFile) {
    fclose(g_hLogicDevs.mRawDataReceiver.pRxFile);
    g_hLogicDevs.mRawDataReceiver.pRxFile = NULL;
  }

  DemoLogf("DemoRawDataReceiver: nEventCount=%d, nReadDataCount=%d, nMaxFrameLen=%d\n", nEventCount, nReadDataCount, nMaxFrameLen);
  DemoLogf("DemoRawDataReceiver: Done (time=%2.1f seconds).\n", (double)(stopTime - startTime) / CLOCKS_PER_SEC);
  
  return 0;

}


/*******************************************************************************
  @func
    Raw Data Sender Thread Function 
*******************************************************************************/
void  *DemoRawDataSender(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;
  OTX_RESULT otxRc;
  char msgStr[256];
  clock_t startTime, stopTime;

  // Discard any events from  event queue (if any)
  do {
    result = OtxDrvGetEventData(g_hLogicDevs.mRawDataSender.hEventQueue, &eventData); 
  } while (OTX_RESULT_CODE(result) == OTX_S_OK);
 
  // Flush (empty) the queue of raw data octets waiting to be transmitted (if any).
  result = OtxRawDataSenderFlush(g_hLogicDevs.mRawDataSender.hHandle); DemoCheckRc(g_hLogicDevs.mRawDataSender.hHandle, result);

  startTime = clock();
  g_hLogicDevs.mRawDataSender.nTaskRef = 1;

  // Stuff the Raw Data Sender device full with data before we start to wait for OTX_RAWDATA_SENDER_EC_BUF_DEPLETED events
  do {
    g_hLogicDevs.mRawDataSender.cbTxBuf = fread(g_hLogicDevs.mRawDataSender.sqTxBuf, sizeof(OTX_UINT8), DEMO_RAW_TX_BUF_SIZE, g_hLogicDevs.mRawDataSender.pTxFile);
  
    otxRc = OtxRawDataSendData(g_hLogicDevs.mRawDataSender.hHandle, g_hLogicDevs.mRawDataSender.sqTxBuf, g_hLogicDevs.mRawDataSender.cbTxBuf, g_hLogicDevs.mRawDataSender.nTaskRef, OTX_TIME_INFINITY);

    if (OTX_RESULT_CODE(otxRc) != OTX_E_BUSY) {
      g_hLogicDevs.mRawDataSender.nTaskRef++;
    }

  } while (g_hLogicDevs.mRawDataSender.cbTxBuf && (OTX_RESULT_CODE(otxRc) != OTX_E_BUSY));

  
  if (g_hLogicDevs.mRawDataSender.cbTxBuf) {
    DemoLogf("DemoRawDataSender: Waiting for OTX_RAWDATA_SENDER_EC_BUF_DEPLETED events to complete the transmission...\n");
  } else {
    DemoLogf("DemoRawDataSender: The entire file fit in the driver buffer. Closing file\n");
    fclose(g_hLogicDevs.mRawDataSender.pTxFile);
    g_hLogicDevs.mRawDataSender.pTxFile = NULL;
  }
  
  
  // Wait for events from the Raw Data Sender
  while (1) {
    OtxDrvWaitForSingleEvent(g_hLogicDevs.mRawDataSender.hNotificationEvent, OTX_TIME_INFINITY);
    do {
      result = OtxDrvGetEventData(g_hLogicDevs.mRawDataSender.hEventQueue, &eventData); DemoCheckRc(g_hLogicDevs.mRawDataSender.hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        switch (eventData.m_eDeviceType) {
        case OTX_LDEVICE_DATA_RAW_SENDER:
          // Event from Raw Data Sender 

          switch (eventData.m_nCode) {
          case OTX_RAWDATA_SENDER_EC_BUF_DEPLETED:
            {
              OTX_RESULT otxRc;

              DemoLogf("Buffer sent (eventData.m_nRequestId = %d)\n", eventData.m_nRequestId);
              
              if (g_hLogicDevs.mRawDataSender.pTxFile) {
                while (g_hLogicDevs.mRawDataSender.cbTxBuf) {
                  otxRc = OtxRawDataSendData(g_hLogicDevs.mRawDataSender.hHandle, g_hLogicDevs.mRawDataSender.sqTxBuf, g_hLogicDevs.mRawDataSender.cbTxBuf, g_hLogicDevs.mRawDataSender.nTaskRef, OTX_TIME_INFINITY);

                  if (OTX_RESULT_CODE(otxRc) == OTX_E_BUSY) {
                    break;
                  } else {
                    g_hLogicDevs.mRawDataSender.nTaskRef++;
                  }

                  g_hLogicDevs.mRawDataSender.cbTxBuf = fread(g_hLogicDevs.mRawDataSender.sqTxBuf, sizeof(OTX_UINT8), DEMO_RAW_TX_BUF_SIZE, g_hLogicDevs.mRawDataSender.pTxFile);
                  
                }
                
                if (!g_hLogicDevs.mRawDataSender.cbTxBuf) {
                  DemoLogf("DemoRawDataSender: EOF reached. Closing file\n");
                  fclose(g_hLogicDevs.mRawDataSender.pTxFile);
                  g_hLogicDevs.mRawDataSender.pTxFile = NULL;
                }
              }
            }
            break;
          
          case OTX_RAWDATA_SENDER_EC_COMPLETED:
            DemoLogf("DemoRawDataSender: Raw Data transmission ended.\n");
            goto END;
            break;
            
          default:
            otxRc = OtxDrvEventCode2String(eventData.m_hDevice, eventData.m_nCode, 256, &msgStr[0]);
            // Unexpected event from this device
            if (OTX_SUCCESS(otxRc)) {
              DemoLogf("DemoRawDataSender: Unexpected Event: %s \n", msgStr);
            } else {
              DemoLogf("DemoRawDataSender: Invalid Event: %ld \n", eventData.m_nCode);
            }
            break;
          }
          break;
          default:
            // Unhandled event
            DemoLogf("DemoRawDataSender: Event from an unknown object 0x%X\n", eventData.m_hDevice);
            break;
        }
        
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

END:
  
  stopTime = clock();
  
  DemoLogf("DemoRawDataSender: Done (time=%2.1f seconds, Last nTaskRef=%d).\n", (double)(stopTime - startTime) / CLOCKS_PER_SEC, g_hLogicDevs.mRawDataSender.nTaskRef-1);
  
  return 0;

}



/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;


  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;


}



/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  DemoLogf("%.8X%.8X ", pEventData->m_nTimeStamp.sUint32.nHigh, pEventData->m_nTimeStamp.sUint32.nLow);

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DSP[%d] Event: %s (m_nParam=0x%X, m_nRequestId=0x%X)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam, pEventData->m_nRequestId);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}


/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->logFileName       = DEMO_DEFAULT_LOG_FILE;
  envP->eLaw              = OTX_DATA_FORMAT_COMPAND8_A_LAW;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType        = OTX_DEVICE_THOR_PCI;

  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {        // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {    // Log file
          envP->logFileName = argv[i]+2;
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '2') {
            envP->nBoardType = OTX_DEVICE_THOR_PCI;
            envP->coDsp      = 4;
          } else if ( *(argv[i]+2) == 'T') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_PLUS;
            envP->coDsp      = 4;
          } else if ( *(argv[i]+2) == 'E') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_EXPRESS;
            envP->coDsp      = 4;
          } else if ( *(argv[i]+2) == '4') {
            envP->nBoardType = OTX_DEVICE_THOR_4_PCI_PLUS;
            envP->coDsp      = 4;
          } else if ( *(argv[i]+2) == '8') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCI;
            envP->coDsp      = 4;
          } else if ( *(argv[i]+2) == 'P') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->coDsp      = 4;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'U') {   // U-law Codec PCM Compression Mode
          envP->eLaw              = OTX_DATA_FORMAT_COMPAND8_U_LAW;
        } else if (*(argv[i]+1) == 'A') {   // A-law Codec PCM Compression Mode
          envP->eLaw              = OTX_DATA_FORMAT_COMPAND8_A_LAW;
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }

}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /T[2|T|E|4|8|P]  : Board Type used for the demo, where\n");
  DemoLogf("                  2 = Use Thor-2-PCI with Vidar-5x4 \n");
  DemoLogf("                  T = Use Thor-2-PCI-Plus with Vidar-5x4 Plus \n");
  DemoLogf("                  E = Use Thor-2-PCI-Express with Vidar-5x4 Plus \n");
  DemoLogf("                  4 = Use Thor-4-PCI-Plus with Vidar-5x4 Plus \n");
  DemoLogf("                  8 = Use Thor-8-PCI with Vidar-5x4 \n");
  DemoLogf("                  P = Use Thor-8-PCI-Plus with Vidar-5x4 Plus \n");
  DemoLogf(" /U           : U-law mode\n");
  DemoLogf(" /A           : A-law mode\n");
  DemoLogf(" /Z<str>      : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}



/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" r   Record voice data from HS0\n");
    DemoLogf(" p   Playback voice data from file to HS0\n");
    break;

  case 'r':
    {
      char sqzFileName[64];
      char sqzReordingLength[64];

      DemoLogf("Voice Recorder. Please enter output file: ");
      OsaConsoleGetStr(sqzFileName);

      if ((g_hLogicDevs.mRawDataReceiver.pRxFile = fopen(sqzFileName, "wb")) == NULL) {
        DemoLogf("Error: Could not open file %s \n", sqzFileName);
        break;
      }

      DemoLogf("Please enter recording length (seconds): ");
      OsaConsoleGetStr(sqzReordingLength);

      g_hLogicDevs.mRawDataReceiver.nMsDuration = (clock_t) strtol(sqzReordingLength, NULL, 0) * 1000;

      DemoLogf("Creating Raw Data Receiver Thread (Recording for %d seconds)\n", g_hLogicDevs.mRawDataReceiver.nMsDuration/1000);
#if 0
      g_hLogicDevs.mRawDataReceiver.hThread = CreateThread(
        NULL,                 // Security attributes
        0,                    // Stack size
        DemoRawDataReceiver,  // pointer to thread function
        NULL,                 // argument for new thread
        0,                    // creation flags
        &g_hLogicDevs.mRawDataReceiver.nThreadId            // pointer to receive thread ID
        );
#else
      OsaThreadStart(DemoRawDataReceiver);
#endif
    }
    break;


  case 'p':
    {
      char sqzFileName[64];

      DemoLogf("Voice Playback. Please enter file to playback: ");
      OsaConsoleGetStr(sqzFileName);

      if ((g_hLogicDevs.mRawDataSender.pTxFile = fopen(sqzFileName, "rb")) == NULL) {
        DemoLogf("Error: Could not open file %s \n", sqzFileName);
        break;
      }
    
      DemoLogf("Creating Raw Data Sender Thread\n");
#if 0
      g_hLogicDevs.mRawDataSender.hThread = CreateThread(
        NULL,                 // Security attributes
        0,                    // Stack size
        DemoRawDataSender,    // pointer to thread function
        NULL,                 // argument for new thread
        0,                    // creation flags
        &g_hLogicDevs.mRawDataSender.nThreadId            // pointer to receive thread ID
        );
#else
      OsaThreadStart(DemoRawDataSender);
#endif

    }
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}





/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

	OsaConsoleInit();

  DemoLogf("T2P5X4RECORDER, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("Raw Data Sender/Receiver example program.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
  	OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
  	OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");

  // Use Internal clock source
  result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_INTERNAL); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling physical devices...\n");
  DemoEnablePhysicalDevices();

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("Connecting devices...\n");
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoLogf("Initializing devices...\n");
  DemoInitializeLogicalDevices();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling logical devices...\n");
  DemoEnableLogicalDevices();

  DemoLogf("Command loop running. Press '?' for help, 'q' to quit\n");

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300);
  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();

  return 0;
}

